'use strict';

const co = require('co');

const chalk = require('chalk');

const FS = require('mz/fs');

const Path = require('path');

const reg = /^([^\@]+)\@(\@?[^\@]+)$/;

const scopeReg = /^\@[^\@]+$/;

const fileLinkReg = /.+\/([^\@\/]+)\@((?:[^\@\/]+)|(?:\@[^\@\/]+\/[^\@\/]+))\/(.+)$/;

const utils = require('./utils');

const spinner = utils.spinner();

const strictPackageNames = [
    'rn-core',
    'react',
    'react-native'
];

module.exports = class {

    constructor(config) {

        this.$config = config || {};
    }

    get $latestVersions() {

        return this.$config.latestVersions || (new Map);
    }

    get $storeDir() {

        return this.$config.storeDir || '';
    }

    * $isDir(path) {

        if (yield FS.exists(path)) {

            const status = yield FS.stat(path);

            return status.isDirectory();
        }
        else {

            return false;
        }
    }

    * $isFile(path) {

        if (yield FS.exists(path)) {

            const status = yield FS.stat(path);

            return status.isFile();
        }
        else {

            return false;
        }
    }

    * $getSoftLink(path) {

        try {

            return yield FS.readlink(path);
        }
        catch (err) {

            return '';
        }
    }

    $isLatestVersion(path) {

        const me = this;

        const pkg = require(Path.join(path, 'package.json'));

        return pkg.version === me.$latestVersions.get(pkg.name);
    }

    * $checkPackageVersions(path) {

        const me = this;

        const name = Path.basename(path);

        if (strictPackageNames.indexOf(name) !== -1) {

            const version = require(Path.join(path, 'package.json')).version;

            const latestVersion = me.$latestVersions.get(name);

            if (version !== latestVersion) {

                console.error(chalk.red(`\nError: ${path}\n  ${name}:\n    Different version included: ${version}\n    Latest version included is: ${latestVersion}`));

                process.exit(1);
            }
        }
    }

    * $handleBinFileSoftLinks(path) {

        const me = this;

        const files = yield FS.readdir(path);

        for (const file of files) {

            const filePath = Path.join(path, file);

            const softLinkPath = yield me.$getSoftLink(filePath);

            if (softLinkPath) {

                const matchResult = softLinkPath.match(fileLinkReg);

                if (matchResult) {

                    const name = matchResult[2], version = matchResult[1].substr(1), endPath = matchResult[3];

                    yield FS.unlink(filePath);

                    if (version === me.$latestVersions.get(name)) {

                        yield FS.symlink(Path.relative(path, Path.join(me.$storeDir, name, endPath)), filePath, 'junction');
                    }
                    else {

                        yield FS.symlink(Path.relative(path, Path.join(path, '..', name, endPath)), filePath, 'junction');
                    }
                }
            }
        }
    }

    * $copy(from, to) {

        const me = this;

        let dirs = yield FS.readdir(from);

        for (const dir of dirs) {

            const fromPath = Path.join(from, dir);

            const toPath = Path.join(to, dir);

            if (yield me.$isDir(fromPath)) {

                let softLinkPath = yield me.$getSoftLink(fromPath);

                if (softLinkPath) {

                    softLinkPath = Path.isAbsolute(softLinkPath) ? softLinkPath : Path.join(from, softLinkPath);

                    yield FS.symlink(Path.relative(to, softLinkPath), toPath, 'junction');
                }
                else {

                    yield FS.mkdir(toPath);

                    yield me.$copy(fromPath, toPath);
                }
            }
            else {

                let softLinkPath = yield me.$getSoftLink(fromPath);

                if (softLinkPath) {

                    softLinkPath = Path.isAbsolute(softLinkPath) ? softLinkPath : Path.join(from, softLinkPath);

                    yield FS.symlink(Path.relative(to, softLinkPath), toPath, 'junction');
                }
                else {

                    yield FS.writeFile(toPath, yield FS.readFile(fromPath));

                    yield FS.chmod(toPath, (yield FS.stat(fromPath)).mode);
                }
            }
        }

        if (Path.basename(to) === 'node_modules') {

            yield me.start(to);

            dirs = yield FS.readdir(to);

            if (!dirs.length) {

                yield utils.rimraf(to);
            }
        }
    }

    * start(baseDir) {

        const me = this;

        baseDir = baseDir || me.$storeDir;

        let isRoot = false;

        if (baseDir === me.$storeDir) {

            isRoot = true;
        }
        else if (scopeReg.test(Path.basename(baseDir)) && Path.join(baseDir, '..') === me.$storeDir) {

            isRoot = true;
        }

        const dirs = yield FS.readdir(baseDir);

        for (const dir of dirs) {

            const path = Path.join(baseDir, dir);

            let softLinkPath = yield me.$getSoftLink(path);

            if (softLinkPath) {

                softLinkPath = Path.isAbsolute(softLinkPath) ? softLinkPath : Path.join(baseDir, softLinkPath);

                yield FS.unlink(path);

                if (isRoot || !me.$isLatestVersion(softLinkPath)) {

                    spinner.text = `Removing soft links: ${dir}`;

                    yield FS.mkdir(path);

                    yield me.$copy(softLinkPath, path);

                    yield me.$checkPackageVersions(path);
                }
            }
            else if (scopeReg.test(dir)) {

                yield me.start(path);

                let dirs = yield FS.readdir(path);

                if (!dirs.length) {

                    yield utils.rimraf(path);
                }
            }
            else if (dir === '.bin') {

                yield me.$handleBinFileSoftLinks(path);
            }
        }
    }

    * clear(baseDir) {

        const me = this;

        baseDir = baseDir || me.$storeDir;

        const dirs = yield FS.readdir(baseDir);

        for (const dir of dirs) {

            if (reg.test(dir)) {

                spinner.text = `Removing redundant packages: ${dir.replace(reg, '$2')}`;

                yield utils.rimraf(Path.join(baseDir, dir));
            }
        }
    }
}
